module AuditModel
  module Models
    module Helpers
      extend ActiveSupport::Concern

      module ClassMethods
        def auditable(options = {}) 
          send :include, InstanceMethods
          class_attribute :audit_enabled, instance_writer: false
          options[:on] ||= [:create, :update, :destroy]
          setup_callbacks options[:on]
          setup_model
          setup_associations

          send :enable_audit
        end

        def without_audit 
          cache = audit_enabled
          disable_audit
          yield
        ensure
          enable_audit if cache
        end

        def disable_audit
          self.audit_enabled = false
        end

        def enable_audit
          self.audit_enabled = true
        end

        private
        def setup_callbacks(options_on = [])
          options_on.each do |option|
            send "callback_on_#{option}"
          end
        end

        def setup_model
          Object.const_set audit_class, Class.new(Models::Base)
        end

        def setup_associations
          has_many :revisions, as: :model, class_name: "AuditModel::Models::Revision"
          has_many :audits, through: :revisions, source: :audit, source_type: audit_class
        end

        def callback_on_create
          after_create :audit_on_create
        end

        def callback_on_update
          before_update :audit_on_update
        end

        def callback_on_destroy
          after_destroy :audit_on_destroy
        end

        def audit_class 
          "#{name}Audit"
        end
      end

      module InstanceMethods
        private

        def audit_on_create
          Revision.audit(self, :create) if self.class.audit_enabled 
        end

        def audit_on_update
          Revision.audit(self, :update) if self.changed? && self.class.audit_enabled
        end

        def audit_on_destroy
          Revision.audit(self, :destroy) unless self.new_record? || !self.class.audit_enabled
        end
        
      end
    end
  end
end